/*
** Module   : XDocumentImpl
** Abstract : DOM XML Document resource type implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130204 Created initial version. Major rework to use w3c backend.
** 002 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 003 CA  20130308 Refactored as XmlFactory APIs now receive a handle instance instead of 
**                  returning a new one.
*/

package com.goldencode.p2j.xml;

import java.io.*;
import org.w3c.dom.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.XmlHelper;

import javax.xml.parsers.ParserConfigurationException;

/**
 * Implementation of the X-Document specific features of the Progress DOM XML
 * support objects.
 */
public class XDocumentImpl
extends XEntityImpl
implements XDocument
{
   /** The real document to assign with handle.  */
   private Document xdoc = null;
   
   /** The root node of the current X-Document. */
   private XNodeRefImpl xRootNode = null;

   /** The encoding current X-Document. */
   private String encoding = null;

   /**
    * Default constructor.
    */
   public XDocumentImpl()
   {
      super();
      setResourceType("X-DOCUMENT");
      try
      {
         xdoc = XmlHelper.newDocument();
      }
      catch (ParserConfigurationException e)
      {
         // throw the error
         String errmsg = e.getMessage();//TODO verifyin 4gl message and code
         ErrorManager.recordOrThrowError(557, errmsg);
         xdoc = null;
      }
      xRootNode = new XNodeRefImpl();

   }

   /**
    * Getting the Element backend object for this XEntity.
    *
    * @return The Element associated with object.
    */
   @Override
   public Node getNode()
   {
      return xdoc;
   }

   /**
    * Reports if this object is unknown.
    *
    * @return <code>true</code> if object is unknown.
    */
   @Override
   public boolean unknown()
   {
      return super.unknown() || xdoc == null;    //TODO autogenerated code by evk,
      // please fix this
   }

   /**
    * Reports if this object is valid for use.
    *
    * @return <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return super.valid() && xdoc != null;//TODO autogenerated code by evk, please fix this
   }

   // Methods
   /**
    * Creating the new node within the current Document object.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNode(handle hChild, String name, String type)
   {
      return createNode(hChild, new character(name), new character(type));
   }

   /**
    * Creating the new node within the current Document object.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNode(handle hChild, character name, String type)
   {
      return createNode(hChild, name, new character(type));
   }

   /**
    * Creating the new node within the current Document object.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNode(handle hChild, String name, character type)
   {
      return createNode(hChild, new character(name), type);
   }

   /**
    * Creating the new node within the current Document object.
    *
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    *
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNode(handle hChild, character name, character type)
   {
      if (type.isUnknown())
      {
         return new logical(false);
      }
      XmlFactory.createXNodeReference(hChild, xdoc, name, type);
      
      return hChild.isValid();
   }

   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, String uri,
                                      String name, String type)
   {
      return createNodeNamespace(hChild, new character(uri),
                                 new character(name), new character(type));
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, character uri,
                                      String name, String type)
   {
      return createNodeNamespace(hChild, uri, new character(name),
                                 new character(type));
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, String uri,
                                      character name, String type)
   {
      return createNodeNamespace(hChild, new character(uri), name,
                                 new character(type));
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, String uri,
                                      String name, character type)
   {
      return createNodeNamespace(hChild, new character(uri),
                                 new character(name), type);
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, String uri,
                                      character name, character type)
   {
      return createNodeNamespace(hChild, new character(uri), name, type);
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, character uri,
                                      String name, character type)
   {
      return createNodeNamespace(hChild, uri, new character(name), type);
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, character uri,
                                      character name, String type)
   {
      return createNodeNamespace(hChild, uri, name, new character(type));
   }
   
   /**
    * Creating the new node within the current Document object using different
    * approach based on the name space URI.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be created and
    *           returned on method call to use further.
    * @param    uri
    *           The URI namespace of the node to be created.
    * @param    name
    *           The name of the node to be created.
    * @param    type
    *           The type of the node to be created.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical createNodeNamespace(handle hChild, character uri,
                                      character name, character type)
   {
      XmlFactory.createXNodeReference(hChild, xdoc, uri, name, type);
      
      return hChild.isValid();
   }

   /**
    * Getting the root node for the current document. The handle is placed
    * into passed object reference.
    * 
    * @param    hRoot
    *           The valid handle class instance variable to receive the
    *           result.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical getDocumentElement(handle hRoot)
   {
      if (unknown() || !valid() )
      {
         return new logical(false);
      }
      hRoot.assign(xRootNode);

      return new logical(valid());
   }

   /**
    * Loading X-Document form file or memptr sources.
    * 
    * @param    mode
    *           Variable indicating the type of source to read. "file" means
    *           reading from file, "memptr" means reading from memory buffer.
    * @param    source
    *           If mode == "file" this is the file name, if "memptr" this is
    *           the memptr object.
    * @param    validate
    *           Flag indicating whether to validate while reading or not.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical load(String mode, Object source, logical validate)
   {
      return load(new character(mode), source, validate);
   }
   
   /**
    * Loading X-Document form file or memptr sources.
    * 
    * @param    mode
    *           Variable indicating the type of source to read. "file" means
    *           reading from file, "memptr" means reading from memory buffer.
    * @param    source
    *           If mode == "file" this is the file name, if "memptr" this is
    *           the memptr object.
    * @param    validate
    *           Flag indicating whether to validate while reading or not.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical load(String mode, Object source, boolean validate)
   {
      return load(new character(mode), source, new logical(validate));
   }
   
   /**
    * Loading X-Document form file or memptr sources.
    * 
    * @param    mode
    *           Variable indicating the type of source to read. "file" means
    *           reading from file, "memptr" means reading from memory buffer.
    * @param    source
    *           If mode == "file" this is the file name, if "memptr" this is
    *           the memptr object.
    * @param    validate
    *           Flag indicating whether to validate while reading or not.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical load(character mode, Object source, boolean validate)
   {
      return load(mode, source, new logical(validate));
   }
   
   /**
    * Loading X-Document form file or memptr sources.
    * 
    * @param    mode
    *           Variable indicating the type of source to read. "file" means
    *           reading from file and source contains file name,
    *           "memptr" means reading from memory buffer.
    * @param    source
    *           If mode == "file" this is the file name, if "memptr" this is
    *           the memptr object.
    * @param    validate
    *           Flag indicating whether to validate while reading or not.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical load(character mode, Object source, logical validate)
   {
      String loadMode = mode.getValue();
      
      if (loadMode.equalsIgnoreCase("file"))
      {
         try
         {
            xdoc = XmlHelper.parse((String)source, validate.getValue(), null);
         }
         catch (Exception e)
         {
            final String errmsg = e.getMessage();//TODO verifyin 4gl message and code
            ErrorManager.recordOrThrowError(557, errmsg);
         }
      }
      else if (loadMode.equalsIgnoreCase("memptr"))
      {
         try
         {
            // TODO review it
            memptr mptr = (memptr)source;
            ByteArrayInputStream bais = new ByteArrayInputStream(mptr.getValue());
            xdoc = XmlHelper.parse(bais, validate.getValue(), null);
         }
         catch (Exception e)
         {
            final String errmsg = e.getMessage();//TODO verifyin 4gl message and code
            ErrorManager.recordOrThrowError(557, errmsg);
         }
      }
      else
      {
         return new logical(false);
      }
      
      return new logical(true);
   }
   
   /**
    * Saving X-Document to file or stream or memptr object.
    * 
    * @param    mode
    *           Variable indicating the type of target to write. "file" means
    *           writing to file, "stream" means writing to the stream and
    *           "memptr" means writing to memory buffer.
    * @param    target
    *           If mode == "file" this is the file name, if "stream" means
    *           output stream, if "memptr" this is the memptr object.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical save(String mode, Object target)
   {
      return save(new character(mode), target);
   }
   
   /**
    * Saving X-Document to file or stream or memptr object.
    * 
    * @param    mode
    *           Variable indicating the type of target to write. "file" means
    *           writing to file, "stream" means writing to the stream and
    *           "memptr" means writing to memory buffer.
    * @param    target
    *           If mode == "file" this is the file name, if "stream" means
    *           output stream, if "memptr" this is the memptr object.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical save(character mode, Object target)
   {
      String writeMode = mode.getValue();
      if (writeMode.equalsIgnoreCase("file"))
      {
         try
         {
            XmlHelper.write(xdoc, (String)target);
         }
         catch (Exception e)
         {
            final String errmsg = e.getMessage();//TODO verifyin 4gl message and code
            ErrorManager.recordOrThrowError(557, errmsg);
         }
      }
      else if (writeMode.equalsIgnoreCase("stream"))
      {
         try
         {
            XmlHelper.write(xdoc, (OutputStream)target);
         }
         catch (Exception e)
         {
            final String errmsg = e.getMessage();//TODO verifyin 4gl message and code
            ErrorManager.recordOrThrowError(557, errmsg);
         }
      }
      else if (writeMode.equalsIgnoreCase("memptr"))
      {
         try
         {
            // TODO implement it
            final memptr memptrTarget = (memptr)target;
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            XmlHelper.write(xdoc, baos);
         }
         catch (Exception e)
         {
            final String errmsg = e.getMessage();//TODO verifyin 4gl message and code
            ErrorManager.recordOrThrowError(557, errmsg);
         }
      }
      else
      {
         return new logical(false);
      }
      
      return new logical(true);
   }

   /**
    * Creates a new XML document, initializes the document based on the referenced DTD, and
    * creates its root node.
    *
    * @param namespaceUri character expression representing the namespace Uniform Resource
    *                     Identifier (URI) you want associated with the root node of the XML
    *                     document. The namespace-uri must be unique and persistent.
    * @param rootNodeName A character expression representing the name of the root node as defined
    *                     in the XML document. If you are using namespaces and you want to
    *                     associate a prefix with the namespace, you must qualify this node name
    *                     with the namespace-uri and a colon character prefix (for example,
    *                     namespace-uri:root-node-name). You must explicitly set the xmlns
    *                     attribute on the root node.
    * @param publicId     An optional character expression representing the public ID of the DTD.
    *                     Currently, there is no way to retrieve a DTD based on a public ID.
    * @param systemId     A required character expression representing the system ID of the DTD.
    *                     This contains the path to the DTD which is either a file system path or
    *                     an HTTP URL. The ABL parser uses this information to retrieve the DTD
    *                     when parsing the document.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical initializeDocumentType(character namespaceUri, character rootNodeName,
                                         character publicId, character systemId)
   {
      try
      {
         xdoc = XmlHelper.newDocument(null, publicId.toString(), systemId.toString());
         String nameImpl = null;
         String nsUriImpl = null;

         if (!namespaceUri.isUnknown())
         {
            nsUriImpl = namespaceUri.toStringMessage();
         }
         if (!rootNodeName.isUnknown())
         {
            nameImpl = rootNodeName.toStringMessage();
         }
         xRootNode = new XNodeRefImpl(xdoc.createElementNS(nsUriImpl, nameImpl));
         getHandle().assign(this);
      }
      catch (ParserConfigurationException e)
      {
         final String errmsg = e.getMessage();//TODO verifyin 4gl message and code
         ErrorManager.recordOrThrowError(667, errmsg);
      }
      return new logical(true);
   }

   /**
    * Get encoding
    *
    * @return encoding of xDoc
    */
   @Override
   public character getEncoding()
   {
      final character result;
      if (encoding != null)
      {
         result = new character(encoding);
      }
      else
      {
         final String xdocEncoding = xdoc.getXmlEncoding();

         if (xdocEncoding == null)
         {
            result = new character("");
         }
         else
         {
            result = new character(xdocEncoding);
         }
      }

      return result;
   }

   /**
    * Get encoding
    *
    */
   @Override
   public void setEncoding(character encoding)
   {
      this.encoding = encoding.toString();//TODO string or character ?
   }

   /**
    * Sets the new value of the encoding attribute for the given SAX-Writer object.
    *
    * @param encoding The encoding value to be set for the given writer.
    */
   @Override
   public void setEncoding(String encoding)
   {
      setEncoding(new character(encoding));
   }

   /**
    * Determines the list of namespace/location pairs of an XML Schema file to validate
    * against.
    *
    * @param    schemaLocation
    *           Determines the list of namespace/location pairs of an XML Schema file to
    *           validate against.
    */
   @Override
   public void setSchemaLocation(character schemaLocation)
   {
      xdoc.getDomConfig().setParameter("schema-location", getSchemaLocation() + " " +
              schemaLocation.toString());
   }

   /**
    * Determines the list of namespace/location pairs of an XML Schema file to validate
    * against.
    *
    * @return   schema location, default value is "".
    */
   @Override
   public character getSchemaLocation()
   {
      final character result;
      final String schemaLocation =
              (String)xdoc.getDomConfig().getParameter("schema-location");
      if (schemaLocation != null) {
         result = new character(schemaLocation);
      } else {
         result = new character("");
      }
      return result;  //TODO
   }

   /**
    * A delimiter-separated list of directory paths used to locate the XML Document Type
    * Definition (DTD) associated with a particular XML document. It is searched if the XML
    * document contains a relative path to the DTD.
    *
    * @param schemaPath To separate directory paths, use a comma, or a delimiter character that
    *                   your operating system recognizes.
    */
   @Override
   public void setSchemaPath(character schemaPath)
   {
      xdoc.getDomConfig().setParameter("schema-location", schemaPath);
      //TODO autogenerated code by evk, please fix this
   }

   /**
    * A delimiter-separated list of directory paths used to locate the XML Document Type
    * Definition (DTD) associated with a particular XML document. It is searched if the XML
    * document contains a relative path to the DTD.
    *
    * @return   schema path, default value is unknown(?).
    */
   @Override
   public character getSchemaPath()
   {
      return new character();  //TODO The default is the Unknown value (?).
   }

   /**
    * This method returns the public ID of the external DTD from which an XML document was
    * generated.
    *
    * @return public-id of DTD
    */
   @Override
   public character getPublicId()
   {
      if (!valid())
      {
         return new character();
      }
      DocumentType documentType = xdoc.getDoctype();
      if (documentType != null)
      {
         return new character(documentType.getPublicId());
      }
      return new character();
   }

   /**
    * Returns the system ID of the external DTD from which an XML document was generated. This
    * contains the path to the DTD which is either a file system path or an HTTP URL. The ABL
    * parser uses this information to retrieve the DTD when parsing the document.
    *
    * @return system-id of DTD
    */
   @Override
   public character getSystemId()
   {
      if (!valid())
      {
         return new character();
      }
      return new character(xdoc.getDoctype().getSystemId());
   }

   /**
    * Creating the new node within the current Document object.
    *
    * @param xNodeRef       The valid handle class instance to use for the new XML node.
    * @param xSourceNodeRef The valid handle class instance to import from.
    * @param deep           A logical that if <code>true</code> specifies that the whole sub-tree is to
    *                       be copied. The default value is {@code false}.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical importNode(handle xNodeRef, handle xSourceNodeRef, logical deep)
   {
      if (unknown() || !valid() )
      {
         return new logical(false);
      }
      final XNodeRefImpl nodeRef = (XNodeRefImpl) xSourceNodeRef.get();
      final Node node = xdoc.importNode(nodeRef.getNode(), deep.booleanValue());
      xNodeRef.assign(new XNodeRefImpl(node));
      return xNodeRef.isValid();
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(String prefix)
   {
      setNamespacePrefix(new character(prefix));
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(character prefix)
   {
      readOnlyError("NAMESPACE-PREFIX");
   }

   /**
    * Read only attribute PUBLIC-ID
    */
   @Override
   public void setPublicId()
   {
      readOnlyError("PUBLIC-ID");
   }

   /**
    * Read only attribute SYSTEM-ID
    */
   @Override
   public void setSystemId()
   {
      readOnlyError("SYSTEM-ID");
   }

   /**
    * Adding new match pair for namespace schema and physical file location.
    *
    * @param    namespace
    *           Target namespace of the schema or am empty string or Unknown value.
    * @param    location
    *           Location of the XML schema file.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical addSchemaLocation(character namespace, character location)
   {
      // TODO: bad place, this method only for XDoc and SaxReader
//      return new logical(true);
//      Namespace XSI = Namespace.getNamespace("xsi", "http://www.w3.org/2001/XMLSchema-instance");
//      root.addNamespaceDeclaration(XSI);
//      xdoc.getDocumentElement().setN
      return new logical(true);

   }

}
